"""This module handles lists"""

import itertools


def de_duplicate(items):
    """Remove any duplicate item, preserving order

    >>> assert de_duplicate([1, 9, 2, 8, 1, 7, 2]) == [1, 9, 2, 8, 7]
    """
    result = []
    for item in items:
        if item not in result:
            result.append(item)
    return result


def flatten(list_of_lists):
    """Reduce a lists of lists to a list, comprehensively

    >>> assert flatten([[1, 2], [3, 4]]) == [1, 2, 3, 4]

    Thanks to Alex Martelli at
        https://stackoverflow.com/a/952952/500942
    """
    return [item for list_ in list_of_lists for item in list_]


def flatten_(list_of_lists):
    """Reduce a lists of lists to a list, functionally

    >>> assert flatten_([[1, 2], [3, 4]]) == [1, 2, 3, 4]

    Thanks to CTT at
        https://stackoverflow.com/a/716482/500942
    """
    return list(itertools.chain.from_iterable(list_of_lists))






"""Handle scripts from the command_line

This module is a simplifying proxy to stdlib's sys.exit()
"""

import inspect
import os
import sys

import stackprinter

from pysyte.cli import arguments
from pysyte.debuggers import DebugExit


def try_main(main):
    """Handle expected exceptions"""
    try:
        return os.EX_OK if main() else os.X_OK
    except KeyboardInterrupt as e:
        sys.stderr.write('^c ^C ^c    ^C ^c ^C    ^c ^C ^c\n')
        ctrl_c = 3
        return ctrl_c
    except DebugExit:
        return os.X_OK
    except SystemExit as e:
        return e.code
    except Exception as e:  # pylint: disable=broad-except
        stackprinter.show(e, style='darkbg')


def run(main_method, add_args=None, post_parse=None):
    """Run a main_method from command line, parsing arguments

    if add_args(parser) is given it should add arguments to the parser
    if post_parse(args) is given then it is called with parsed args
        if it returns the args they will be used
    Then call:
        main_method(args)
    """

    module = inspect.getmodule(main_method)
    if not add_args:
        main = main_method
        assert not main_method.__code__.co_argcount
    else:
        def main():
            args = parser.parse_args()
            if post_parse:
                parsed = post_parse(args)
                if parsed:
                    args = parsed
            return main_method(args)

        parser = arguments.parser(module.__doc__)
        add_args(parser)

    if module.__name__ == '__main__':
        sys.exit(try_main(main))



"""Handle arguments from the command line

This module is a simplifying proxy to stdlib's argparse
"""

import argparse
import sys
from functools import partial
from itertools import chain


def extract_strings(names, name):
    result = {}
    try:
        value = names[name]
    except KeyError:
        return []
    if isinstance(value, str):
        return [value]
    try:
        return [v for v in value if isinstance(v, str)]
    except TypeError:
        return []


def parser(description=None, usage=None):
    """Make a command line argument parser"""

    class ArgumentsNamespace(object):
        def __init__(self, result):
            self._result = result

        def __getattr__(self, name):
            try:
                super(ArgumentsNamespace, self).__getattr__(name)
            except AttributeError:
                return getattr(self._result, name)

        def get_arg(self, name):
            if not self._result:
                return None
            return getattr(self._result, name)

        def get_strings(self, name):
            if not self._result:
                return None
            return extract_strings(self._result.__dict__, name)

    class ArgumentsParser(object):
        def __init__(self, usage, epilog):
            self.parser = argparse.ArgumentParser(usage=usage, epilog=epilog)
            self._parsed = None

            self.parse = self.parse_args

            self.string = self.arg = self.parser.add_argument
            self.boolean = self.opt = self.true = partial(self.arg, action='store_true')
            self.integer = self.int = partial(self.arg, type=int)
            self.strings = partial(self.arg, type=str, nargs='*')

        def __repr__(self):
            return f'<{self.__class__.__name__}>'

        def positional(self, *args, **kwargs):
            """Add optional positional args"""
            return self.string(*args, **kwargs, nargs='*')

        def positionals(self, *args, **kwargs):
            """Add mandatory positional args"""
            return self.string(*args, **kwargs, nargs='+')

        def parse_args(self, args=None):
            return ArgumentsNamespace(self.parser.parse_args(args))


    lines = description.splitlines()
    epilog = None
    if not usage:
        if len(lines) > 1 and not lines[1]:
            usage, epilog = lines[0], '\n\n'.join(lines[2:])
    else:
        usage = description = lines[0]
    return ArgumentsParser(usage, epilog)
